from collections import defaultdict
from itertools import product
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Union
import yaml

from opa_tools.config import REQUEST_DIR


def _get_single_request(num: int, kind: str) -> Dict:
    if kind not in ("opa", "gsv"):
        raise AssertionError(
            f"kind must be either 'opa' or 'gsv', but was '{kind}'."
        )
    request_path = REQUEST_DIR / f"{num:02d}" / f"{kind}.yml"
    with open(request_path, "r", encoding="utf-8") as rf:
        return yaml.safe_load(rf)

def process_dct(dct, prefix="", valid_types=(str, int, float), force_fun=str, wrapper_fun=None):
    processed_dict = {}
    for key, value in dct.items():
        processed_key = f"{prefix}{key}"
        if valid_types and force_fun:
            processed_value = value if isinstance(value, valid_types) else force_fun(value)
        else:
            processed_value = value
        if wrapper_fun:
            processed_value = wrapper_fun(processed_value)
        processed_dict[processed_key] = processed_value
    return processed_dict


def get_request(num: Optional[int] = None, /, kind: Optional[str] = None, *, single_as_list=False) -> Union[List[Dict], Dict]:
    if num is None:
        output = []
        for num in get_request_names():
            output.extend(get_request(num, kind=kind, single_as_list=True))
        return output
    if kind is None:
        return [
            _get_single_request(num=num, kind="gsv"),
            _get_single_request(num=num, kind="opa"),
        ]
    if single_as_list:
        return [_get_single_request(num=num, kind=kind)]
    return _get_single_request(num=num, kind=kind)


def get_request_names():
    return sorted(int(path.name) for path in REQUEST_DIR.iterdir())

def request_product(infile: Path, outpath: Optional[Path] = None):
    with open(infile, "r") as rf:
        req_items = list(yaml.safe_load(rf).items())
    requests = []
    req_keys, req_values = zip(*req_items)
    for value_comb in product(*req_values):
        requests.append(dict(zip(req_keys, value_comb)))
    return requests


def request_wrap(nums: Sequence[int], /, outfile: Optional[Path] = None):
    req_dcts = defaultdict(set)
    for num in nums:
        gsvrequest, oparequest = get_request(num)
        request = {
            **process_dct(gsvrequest, prefix="gsv_"),
            **process_dct(oparequest, prefix="opa_"),
        }
        print(request)
        for key, value in request.items():
            # Making hashable in a cheap way by turning into string
            req_dcts[key].add(value)
    req_dcts = {key: list(values) for key, values in req_dcts.items()}
    if outfile:
        with open(outfile, "w") as wf:
            yaml.dump(req_dcts, wf)
    return req_dcts

